from .ridgeClassifier import RidgeClassifier
from .ridgeRegressor import RidgeRegressor
from .ridgeClassifierMtask import RidgeClassifierMtask

__all__ = ["RidgeClassifier", "RidgeRegressor", "RidgeClassifierMtask"]
